# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.shared_file_system.v2 import share_replica as _share_replica
from openstack.tests.functional.shared_file_system.v2 import base


class ShareReplicaTest(base.BaseSharedFileSystemTest):
    min_microversion = '2.7'

    def setUp(self):
        super().setUp()

        self.SHARE_NAME = self.getUniqueString()
        my_share = self.create_share(
            name=self.SHARE_NAME,
            size=2,
            share_type="dhss_false",
            share_protocol='NFS',
            description=None,
        )
        self.user_cloud.shared_file_system.wait_for_status(
            my_share,
            status='available',
            failures=['error'],
            interval=5,
            wait=self._wait_for_timeout,
        )
        self.SHARE = my_share
        self.SHARE_ID = my_share.id
        srt = self.user_cloud.shared_file_system.create_share_replica(
            my_share.id
        )
        self.SHARE_REPLICA = srt
        self.user_cloud.shared_file_system.wait_for_status(
            self.SHARE_REPLICA,
            status='in_sync',
            failures=['error'],
            interval=5,
            wait=self._wait_for_timeout,
            attribute='replica_state',
        )
        self.assertIsNotNone(srt)
        self.assertIsNotNone(srt.id)
        self.SHARE_REPLICA_ID = srt.id

    def tearDown(self):
        share_replicas = self.user_cloud.share.share_replicas()
        for share_replica in share_replicas:
            replica_state = getattr(share_replica, 'replica_state')
            if replica_state != 'active':
                self.user_cloud.shared_file_system.delete_share_replica(
                    share_replica, ignore_missing=True
                )
                self.user_cloud.shared_file_system.wait_for_delete(
                    share_replica, interval=2, wait=self._wait_for_timeout
                )
        sot = self.user_cloud.shared_file_system.delete_share(
            self.SHARE_ID, ignore_missing=True
        )
        self.user_cloud.shared_file_system.wait_for_delete(
            self.SHARE, interval=2, wait=self._wait_for_timeout
        )
        self.assertIsNone(sot)
        super().tearDown()

    def test_get(self):
        sot = self.user_cloud.shared_file_system.get_share_replica(
            self.SHARE_REPLICA_ID
        )
        assert isinstance(sot, _share_replica.ShareReplica)
        self.assertEqual(self.SHARE_REPLICA_ID, sot.id)

    def test_list_share_replicas(self):
        share_replicas = self.user_cloud.share.share_replicas()
        self.assertGreater(len(list(share_replicas)), 0)
        for share_replica in share_replicas:
            for attribute in ('id', 'name', 'created_at', 'replica_state'):
                self.assertTrue(hasattr(share_replica, attribute))

    def test_reset(self):
        res = self.operator_cloud.share.reset_share_replica_status(
            self.SHARE_REPLICA_ID, 'error'
        )
        self.assertIsNone(res)
        sot = self.user_cloud.share.get_share_replica(self.SHARE_REPLICA_ID)
        self.assertEqual('error', sot.status)

    def test_reset_state(self):
        res = self.operator_cloud.share.reset_share_replica_state(
            self.SHARE_REPLICA_ID, 'error'
        )
        self.assertIsNone(res)
        sot = self.user_cloud.share.get_share_replica(self.SHARE_REPLICA_ID)
        self.assertEqual('error', sot.replica_state)

    def test_promote_share_replica(self):
        psr = self.user_cloud.shared_file_system.promote_share_replica(
            self.SHARE_REPLICA_ID
        )
        self.assertIsNone(psr)
        sot = self.user_cloud.share.get_share_replica(self.SHARE_REPLICA_ID)
        self.user_cloud.shared_file_system.wait_for_status(
            sot,
            status='available',
            failures=['error'],
            interval=5,
            wait=self._wait_for_timeout,
        )
        sot = self.user_cloud.share.get_share_replica(self.SHARE_REPLICA_ID)
        self.assertEqual('active', sot.replica_state)

    def test_share_replica_resync(self):
        psr = self.operator_cloud.shared_file_system.resync_share_replica(
            self.SHARE_REPLICA_ID
        )
        self.assertIsNone(psr)
        self.user_cloud.shared_file_system.wait_for_status(
            self.SHARE_REPLICA,
            status='in_sync',
            failures=['error'],
            interval=5,
            wait=self._wait_for_timeout,
            attribute='replica_state',
        )



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from openstack import resource


class QoSMinimumPacketRateRule(resource.Resource):
    resource_key = 'minimum_packet_rate_rule'
    resources_key = 'minimum_packet_rate_rules'
    base_path = '/qos/policies/%(qos_policy_id)s/minimum_packet_rate_rules'

    _allow_unknown_attrs_in_body = True

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True

    # Properties
    #: Traffic direction from the project point of view. Valid values: ('any',
    #: 'egress', 'ingress')
    direction = resource.Body('direction')
    #: Minimum packet rate in kpps.
    min_kpps = resource.Body('min_kpps')
    #: The ID of the QoS policy who owns rule.
    qos_policy_id = resource.URI('qos_policy_id')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1 import adapter

from openstack import exceptions
from openstack import resource
from openstack import utils


class User(resource.Resource):
    resource_key = 'user'
    resources_key = 'users'
    base_path = '/users'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True
    commit_method = 'PATCH'

    _query_mapping = resource.QueryParameters(
        'domain_id',
        'name',
        'password_expires_at',
        is_enabled='enabled',
    )

    # Properties
    #: References the user's default project ID against which to authorize,
    #: if the API user does not explicitly specify one when creating a token.
    #: Setting this attribute does not grant any actual authorization on the
    #: project, and is merely provided for the user's convenience.
    #: Therefore, the referenced project does not need to exist within the
    #: user's domain.
    #:
    #: *New in version 3.1* If the user does not have authorization to
    #: their default project, the default project will be ignored at token
    #: creation. *Type: string*
    default_project_id = resource.Body('default_project_id')
    #: The description of this user. *Type: string*
    description = resource.Body('description')
    #: References the domain ID which owns the user; if a domain ID is not
    #: specified by the client, the Identity service implementation will
    #: default it to the domain ID to which the client's token is scoped.
    #: *Type: string*
    domain_id = resource.Body('domain_id')
    #: The email of this user. *Type: string*
    email = resource.Body('email')
    #: Setting this value to ``False`` prevents the user from authenticating or
    #: receiving authorization. Additionally, all pre-existing tokens held by
    #: the user are immediately invalidated. Re-enabling a user does not
    #: re-enable pre-existing tokens. *Type: bool*
    is_enabled = resource.Body('enabled', type=bool)
    #: The links for the user resource.
    links = resource.Body('links')
    #: Unique user name, within the owning domain. *Type: string*
    name = resource.Body('name')
    #: The default form of credential used during authentication.
    #: *Type: string*
    password = resource.Body('password')
    #: The date and time when the password expires. The time zone is UTC.
    #: A None value means the password never expires.
    #: This is a response object attribute, not valid for requests.
    #: *New in version 3.7*
    password_expires_at = resource.Body('password_expires_at')
    #: A dictionary of users extra options.
    options = resource.Body('options', type=dict, default={})

    def update_password(
        self,
        session: adapter.Adapter,
        current_password: str,
        password: str,
    ) -> None:
        """Update the password for this user.

        :param session: The session to use for the request.
        :param str current_password: The user's current password.
        :param str password: The user's new password.
        """
        url = utils.urljoin(self.base_path, self.id, 'password')
        body = {
            'user': {
                'password': password,
                'original_password': current_password,
            }
        }
        resp = session.post(url, json=body)
        exceptions.raise_from_response(resp)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from keystoneauth1 import adapter

from openstack import exceptions
from openstack import resource
from openstack import utils


class ServerGroup(resource.Resource):
    resource_key = 'server_group'
    resources_key = 'server_groups'
    base_path = '/os-server-groups'

    _query_mapping = resource.QueryParameters("all_projects")

    _max_microversion = '2.64'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_delete = True
    allow_list = True

    # Properties
    #: A name identifying the server group
    name = resource.Body('name')
    #: The list of policies supported by the server group (till 2.63)
    policies = resource.Body('policies')
    #: The policy field represents the name of the policy (from 2.64)
    policy = resource.Body('policy')
    #: The list of members in the server group
    member_ids = resource.Body('members')
    #: The metadata associated with the server group. This is always empty and
    #: only used for preserving compatibility.
    metadata = resource.Body('metadata')
    #: The project ID who owns the server group.
    project_id = resource.Body('project_id')
    #: The rules field, which is a dict, can be applied to the policy.
    #: Currently, only the max_server_per_host rule is supported for the
    #: anti-affinity policy. The max_server_per_host rule allows specifying how
    #: many members of the anti-affinity group can reside on the same compute
    #: host. If not specified, only one member from the same anti-affinity
    #: group can reside on a given host.
    rules = resource.Body('rules', type=dict)
    #: The user ID who owns the server group
    user_id = resource.Body('user_id')

    @classmethod
    def _transform_create_request(
        cls,
        session: adapter.Adapter,
        request: resource._Request,
        *,
        microversion: str | None,
    ) -> resource._Request:
        assert isinstance(request.body, dict)  # narrow type
        # `policy` and `rules` were added with mv=2.64; `policies` was removed.
        body = request.body.get('server_group', {})
        if utils.supports_microversion(session, '2.64'):
            if body.get('policies'):
                if not body.get('policy') and isinstance(
                    body['policies'], list
                ):
                    body['policy'] = body['policies'][0]
                body.pop('policies', None)
        else:  # microversion < 2.64
            if body.get('rules'):
                raise exceptions.NotSupported(
                    "API version 2.64 is required to set rules, but "
                    "it is not available."
                )
            if body.get('policy'):
                if not body.get('policies'):
                    body['policies'] = [body['policy']]
                body.pop('policy', None)
        return request



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from openstack.cloud import _utils
from openstack.cloud import openstackcloud
from openstack.dns.v2 import zone as _zone
from openstack import exceptions


class DnsCloudMixin(openstackcloud._OpenStackCloudMixin):
    def list_zones(self, filters=None):
        """List all available zones.

        :returns: A list of zones dicts.

        """
        if not filters:
            filters = {}
        return list(self.dns.zones(allow_unknown_params=True, **filters))

    def get_zone(self, name_or_id, filters=None):
        """Get a zone by name or ID.

        :param name_or_id: Name or ID of the zone
        :param filters:
            A dictionary of meta data to use for further filtering

        :returns:  A zone dict or None if no matching zone is found.

        """
        if not filters:
            filters = {}
        return self.dns.find_zone(
            name_or_id=name_or_id, ignore_missing=True, **filters
        )

    def search_zones(self, name_or_id=None, filters=None):
        zones = self.list_zones(filters)
        return _utils._filter_list(zones, name_or_id, filters)

    def create_zone(
        self,
        name,
        zone_type=None,
        email=None,
        description=None,
        ttl=None,
        masters=None,
    ):
        """Create a new zone.

        :param name: Name of the zone being created.
        :param zone_type: Type of the zone (primary/secondary)
        :param email: Email of the zone owner (only
            applies if zone_type is primary)
        :param description: Description of the zone
        :param ttl: TTL (Time to live) value in seconds
        :param masters: Master nameservers (only applies
            if zone_type is secondary)

        :returns: a dict representing the created zone.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """

        # We capitalize in case the user passes time in lowercase, as
        # designate call expects PRIMARY/SECONDARY
        if zone_type is not None:
            zone_type = zone_type.upper()
            if zone_type not in ('PRIMARY', 'SECONDARY'):
                raise exceptions.SDKException(
                    f"Invalid type {zone_type}, valid choices are PRIMARY or "
                    f"SECONDARY"
                )

        zone = {
            "name": name,
            "email": email,
            "description": description,
        }
        if ttl is not None:
            zone["ttl"] = ttl

        if zone_type is not None:
            zone["type"] = zone_type

        if masters is not None:
            zone["masters"] = masters

        try:
            return self.dns.create_zone(**zone)
        except exceptions.SDKException:
            raise exceptions.SDKException(f"Unable to create zone {name}")

    @_utils.valid_kwargs('email', 'description', 'ttl', 'masters')
    def update_zone(self, name_or_id, **kwargs):
        """Update a zone.

        :param name_or_id: Name or ID of the zone being updated.
        :param email: Email of the zone owner (only
            applies if zone_type is primary)
        :param description: Description of the zone
        :param ttl: TTL (Time to live) value in seconds
        :param masters: Master nameservers (only applies
            if zone_type is secondary)

        :returns: a dict representing the updated zone.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        zone = self.get_zone(name_or_id)
        if not zone:
            raise exceptions.SDKException(f"Zone {name_or_id} not found.")

        return self.dns.update_zone(zone['id'], **kwargs)

    def delete_zone(self, name_or_id):
        """Delete a zone.

        :param name_or_id: Name or ID of the zone being deleted.

        :returns: True if delete succeeded, False otherwise.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """

        zone = self.dns.find_zone(name_or_id, ignore_missing=True)
        if not zone:
            self.log.debug("Zone %s not found for deleting", name_or_id)
            return False

        self.dns.delete_zone(zone)

        return True

    def list_recordsets(self, zone):
        """List all available recordsets.

        :param zone: Name, ID or :class:`openstack.dns.v2.zone.Zone` instance
            of the zone managing the recordset.

        :returns: A list of recordsets.

        """
        if isinstance(zone, _zone.Zone):
            zone_obj = zone
        else:
            zone_obj = self.get_zone(zone)
        if zone_obj is None:
            raise exceptions.SDKException(f"Zone {zone} not found.")
        return list(self.dns.recordsets(zone_obj))

    def get_recordset(self, zone, name_or_id):
        """Get a recordset by name or ID.

        :param zone: Name, ID or :class:`openstack.dns.v2.zone.Zone` instance
            of the zone managing the recordset.
        :param name_or_id: Name or ID of the recordset

        :returns:  A recordset dict or None if no matching recordset is
            found.

        """
        if isinstance(zone, _zone.Zone):
            zone_obj = zone
        else:
            zone_obj = self.get_zone(zone)
        if not zone_obj:
            raise exceptions.SDKException(f"Zone {name_or_id} not found.")
        return self.dns.find_recordset(
            zone=zone_obj, name_or_id=name_or_id, ignore_missing=True
        )

    def search_recordsets(self, zone, name_or_id=None, filters=None):
        recordsets = self.list_recordsets(zone=zone)
        return _utils._filter_list(recordsets, name_or_id, filters)

    def create_recordset(
        self, zone, name, recordset_type, records, description=None, ttl=None
    ):
        """Create a recordset.

        :param zone: Name, ID or :class:`openstack.dns.v2.zone.Zone` instance
            of the zone managing the recordset.
        :param name: Name of the recordset
        :param recordset_type: Type of the recordset
        :param records: List of the recordset definitions
        :param description: Description of the recordset
        :param ttl: TTL value of the recordset

        :returns: a dict representing the created recordset.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        if isinstance(zone, _zone.Zone):
            zone_obj = zone
        else:
            zone_obj = self.get_zone(zone)
        if not zone_obj:
            raise exceptions.SDKException(f"Zone {zone} not found.")

        # We capitalize the type in case the user sends in lowercase
        recordset_type = recordset_type.upper()

        body = {'name': name, 'type': recordset_type, 'records': records}

        if description:
            body['description'] = description

        if ttl:
            body['ttl'] = ttl

        return self.dns.create_recordset(zone=zone_obj, **body)

    @_utils.valid_kwargs('description', 'ttl', 'records')
    def update_recordset(self, zone, name_or_id, **kwargs):
        """Update a recordset.

        :param zone: Name, ID or :class:`openstack.dns.v2.zone.Zone` instance
            of the zone managing the recordset.
        :param name_or_id: Name or ID of the recordset being updated.
        :param records: List of the recordset definitions
        :param description: Description of the recordset
        :param ttl: TTL (Time to live) value in seconds of the recordset

        :returns: a dict representing the updated recordset.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """

        rs = self.get_recordset(zone, name_or_id)
        if not rs:
            raise exceptions.SDKException(f"Recordset {name_or_id} not found.")

        rs = self.dns.update_recordset(recordset=rs, **kwargs)

        return rs

    def delete_recordset(self, zone, name_or_id):
        """Delete a recordset.

        :param zone: Name, ID or :class:`openstack.dns.v2.zone.Zone` instance
            of the zone managing the recordset.
        :param name_or_id: Name or ID of the recordset being deleted.

        :returns: True if delete succeeded, False otherwise.
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """

        recordset = self.get_recordset(zone, name_or_id)
        if not recordset:
            self.log.debug("Recordset %s not found for deleting", name_or_id)
            return False

        self.dns.delete_recordset(recordset, ignore_missing=False)

        return True



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack.compute.v2 import server_migration
from openstack.tests.unit import base

EXAMPLE = {
    'id': 4,
    'server_id': '4cfba335-03d8-49b2-8c52-e69043d1e8fe',
    'server_uuid': '4cfba335-03d8-49b2-8c52-e69043d1e8fe',
    'user_id': '8dbaa0f0-ab95-4ffe-8cb4-9c89d2ac9d24',
    'project_id': '5f705771-3aa9-4f4c-8660-0d9522ffdbea',
    'created_at': '2016-01-29T13:42:02.000000',
    'updated_at': '2016-01-29T13:42:02.000000',
    'status': 'migrating',
    'source_compute': 'compute1',
    'source_node': 'node1',
    'dest_host': '1.2.3.4',
    'dest_compute': 'compute2',
    'dest_node': 'node2',
    'memory_processed_bytes': 12345,
    'memory_remaining_bytes': 111111,
    'memory_total_bytes': 123456,
    'disk_processed_bytes': 23456,
    'disk_remaining_bytes': 211111,
    'disk_total_bytes': 234567,
}


class TestServerMigration(base.TestCase):
    def setUp(self):
        super().setUp()
        self.resp = mock.Mock()
        self.resp.body = None
        self.resp.json = mock.Mock(return_value=self.resp.body)
        self.resp.status_code = 200
        self.sess = mock.Mock()
        self.sess.post = mock.Mock(return_value=self.resp)

    def test_basic(self):
        sot = server_migration.ServerMigration()
        self.assertEqual('migration', sot.resource_key)
        self.assertEqual('migrations', sot.resources_key)
        self.assertEqual('/servers/%(server_id)s/migrations', sot.base_path)
        self.assertFalse(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_list)
        self.assertFalse(sot.allow_commit)
        self.assertTrue(sot.allow_delete)

    def test_make_it(self):
        sot = server_migration.ServerMigration(**EXAMPLE)
        self.assertEqual(EXAMPLE['id'], sot.id)
        # FIXME(stephenfin): This conflicts since there is a server ID in the
        # URI *and* in the body. We need a field that handles both or we need
        # to use different names.
        # self.assertEqual(EXAMPLE['server_uuid'], sot.server_id)
        self.assertEqual(EXAMPLE['user_id'], sot.user_id)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)
        self.assertEqual(EXAMPLE['created_at'], sot.created_at)
        self.assertEqual(EXAMPLE['updated_at'], sot.updated_at)
        self.assertEqual(EXAMPLE['status'], sot.status)
        self.assertEqual(EXAMPLE['source_compute'], sot.source_compute)
        self.assertEqual(EXAMPLE['source_node'], sot.source_node)
        self.assertEqual(EXAMPLE['dest_host'], sot.dest_host)
        self.assertEqual(EXAMPLE['dest_compute'], sot.dest_compute)
        self.assertEqual(EXAMPLE['dest_node'], sot.dest_node)
        self.assertEqual(
            EXAMPLE['memory_processed_bytes'],
            sot.memory_processed_bytes,
        )
        self.assertEqual(
            EXAMPLE['memory_remaining_bytes'],
            sot.memory_remaining_bytes,
        )
        self.assertEqual(EXAMPLE['memory_total_bytes'], sot.memory_total_bytes)
        self.assertEqual(
            EXAMPLE['disk_processed_bytes'],
            sot.disk_processed_bytes,
        )
        self.assertEqual(
            EXAMPLE['disk_remaining_bytes'],
            sot.disk_remaining_bytes,
        )
        self.assertEqual(EXAMPLE['disk_total_bytes'], sot.disk_total_bytes)

    @mock.patch.object(
        server_migration.ServerMigration,
        '_get_session',
        lambda self, x: x,
    )
    def test_force_complete(self):
        sot = server_migration.ServerMigration(**EXAMPLE)

        self.assertIsNone(sot.force_complete(self.sess))

        url = 'servers/{}/migrations/{}/action'.format(
            EXAMPLE['server_id'],
            EXAMPLE['id'],
        )
        body = {'force_complete': None}
        self.sess.post.assert_called_with(
            url,
            microversion=mock.ANY,
            json=body,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.identity.v3 import project as _project
from openstack.tests.functional.identity.v3 import base


class TestProject(base.BaseIdentityTest):
    def setUp(self):
        super().setUp()

        self.project_name = self.getUniqueString('project')
        self.project_description = self.getUniqueString('project')

    def _delete_project(self, project):
        ret = self.admin_identity_client.delete_project(project)
        self.assertIsNone(ret)

    def test_project(self):
        # create the project

        project = self.admin_identity_client.create_project(
            name=self.project_name,
        )
        self.assertIsInstance(project, _project.Project)
        self.assertEqual('', project.description)
        self.addCleanup(self._delete_project, project)

        # update the project

        project = self.admin_identity_client.update_project(
            project, description=self.project_description
        )
        self.assertIsInstance(project, _project.Project)
        self.assertEqual(self.project_description, project.description)

        # retrieve details of the (updated) project by ID

        project = self.admin_identity_client.get_project(project.id)
        self.assertIsInstance(project, _project.Project)
        self.assertEqual(self.project_description, project.description)

        # retrieve details of the (updated) project by name

        project = self.admin_identity_client.find_project(
            project.name, ignore_missing=False
        )
        self.assertIsInstance(project, _project.Project)
        self.assertEqual(self.project_description, project.description)

        # list all projects

        projects = list(self.admin_identity_client.projects())
        self.assertIsInstance(projects[0], _project.Project)
        self.assertIn(
            self.project_name,
            {x.name for x in projects},
        )

    def test_user_project(self):
        # list all user projects

        assert self.operator_cloud.current_user_id is not None
        user_projects = list(
            self.admin_identity_client.user_projects(
                self.operator_cloud.current_user_id
            )
        )
        self.assertIsInstance(user_projects[0], _project.UserProject)
        self.assertIn(
            self.operator_cloud.current_project_id,
            {x.id for x in user_projects},
        )



# Copyright (c) 2015 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import sys
from typing import TextIO


def setup_logging(
    name: str,
    handlers: list[logging.Handler] | None = None,
    level: int | None = None,
) -> logging.Logger:
    """Set up logging for a named logger.

    Gets and initializes a named logger, ensuring it at least has a
    `logging.NullHandler` attached.

    :param str name:
      Name of the logger.
    :param list handlers:
      A list of `logging.Handler` objects to attach to the logger.
    :param int level:
      Log level to set the logger at.

    :returns: A `logging.Logger` object that can be used to emit log messages.
    """
    handlers = handlers or []
    log = logging.getLogger(name)
    if len(log.handlers) == 0 and not handlers:
        log.addHandler(logging.NullHandler())
    for h in handlers:
        log.addHandler(h)
    if level:
        log.setLevel(level)
    return log


def enable_logging(
    debug: bool = False,
    http_debug: bool = False,
    path: str | None = None,
    stream: TextIO | None = None,
    format_stream: bool = False,
    format_template: str = '%(asctime)s %(levelname)s: %(name)s %(message)s',
    handlers: list[logging.Handler] | None = None,
) -> None:
    """Enable logging output.

    Helper function to enable logging. This function is available for
    debugging purposes and for folks doing simple applications who want an
    easy 'just make it work for me'. For more complex applications or for
    those who want more flexibility, the standard library ``logging`` package
    will receive these messages in any handlers you create.

    :param bool debug:
        Set this to ``True`` to receive debug messages.
    :param bool http_debug:
        Set this to ``True`` to receive debug messages including
        HTTP requests and responses. This implies ``debug=True``.
    :param str path:
        If a *path* is specified, logging output will written to that file
        in addition to sys.stderr.
        The path is passed to logging.FileHandler, which will append messages
        the file (and create it if needed).
    :param stream:
        One of ``None `` or ``sys.stdout`` or ``sys.stderr``.
        If it is ``None``, nothing is logged to a stream.
        If it isn't ``None``, console output is logged to this stream.
    :param bool format_stream:
        If format_stream is False, the default, apply ``format_template`` to
        ``path`` but not to ``stream`` outputs. If True, apply
        ``format_template`` to ``stream`` outputs as well.
    :param str format_template:
        Template to pass to :class:`logging.Formatter`.

    :rtype: None
    """
    if not stream and not path:
        stream = sys.stderr

    if http_debug:
        debug = True
    if debug:
        level = logging.DEBUG
    else:
        level = logging.INFO

    formatter = logging.Formatter(format_template)

    if handlers:
        for handler in handlers:
            handler.setFormatter(formatter)

    else:
        handlers = []

        if stream is not None:
            console = logging.StreamHandler(stream)
            if format_stream:
                console.setFormatter(formatter)
            handlers.append(console)

        if path is not None:
            file_handler = logging.FileHandler(path)
            file_handler.setFormatter(formatter)
            handlers.append(file_handler)

    setup_logging('openstack', handlers=handlers, level=level)
    setup_logging('keystoneauth', handlers=handlers, level=level)

    # Turn off logging on these so that if loggers higher in the tree
    # are more verbose we only get what we want out of the SDK. This is
    # particularly useful when combined with tools like ansible which set
    # debug logging level at the logging root.
    # If more complex logging is desired including stevedore debug logging,
    # enable_logging should not be used and instead python logging should
    # be configured directly.
    setup_logging(
        'urllib3', handlers=[logging.NullHandler()], level=logging.INFO
    )
    setup_logging(
        'stevedore', handlers=[logging.NullHandler()], level=logging.INFO
    )
    # Suppress warning about keystoneauth loggers
    setup_logging('keystoneauth.discovery')
    setup_logging('keystoneauth.identity.base')
    setup_logging('keystoneauth.identity.generic.base')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, cast

from keystoneauth1 import adapter

from openstack.clustering.v1 import _async_resource
from openstack.common import metadata
from openstack import resource
from openstack import utils


class Cluster(_async_resource.AsyncResource, metadata.MetadataMixin):
    resource_key = 'cluster'
    resources_key = 'clusters'
    base_path = '/clusters'

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True
    commit_method = 'PATCH'

    _query_mapping = resource.QueryParameters(
        'name', 'status', 'sort', 'global_project'
    )

    # Properties
    #: The name of the cluster.
    name = resource.Body('name')
    #: The ID of the profile used by this cluster.
    profile_id = resource.Body('profile_id')
    #: The ID of the user who created this cluster, thus the owner of it.
    user_id = resource.Body('user')
    #: The ID of the project this cluster belongs to.
    project_id = resource.Body('project')
    #: The domain ID of the cluster owner.
    domain_id = resource.Body('domain')
    #: Timestamp of when the cluster was initialized.
    #: *Type: datetime object parsed from ISO 8601 formatted string*
    init_at = resource.Body('init_at')
    #: Timestamp of when the cluster was created.
    #: *Type: datetime object parsed from ISO 8601 formatted string*
    created_at = resource.Body('created_at')
    #: Timestamp of when the cluster was last updated.
    #: *Type: datetime object parsed from ISO 8601 formatted string*
    updated_at = resource.Body('updated_at')
    #: Lower bound (inclusive) for the size of the cluster.
    min_size = resource.Body('min_size', type=int)
    #: Upper bound (inclusive) for the size of the cluster. A value of
    #: -1 indicates that there is no upper limit of cluster size.
    max_size = resource.Body('max_size', type=int)
    #: Desired capacity for the cluster. A cluster would be created at the
    #: scale specified by this value.
    desired_capacity = resource.Body('desired_capacity', type=int)
    #: Default timeout (in seconds) for cluster operations.
    timeout = resource.Body('timeout')
    #: A string representation of the cluster status.
    status = resource.Body('status')
    #: A string describing the reason why the cluster in current status.
    status_reason = resource.Body('status_reason')
    #: A dictionary configuration for cluster.
    config = resource.Body('config', type=dict)
    #: A collection of key-value pairs that are attached to the cluster.
    metadata = resource.Body('metadata', type=dict)
    #: A dictionary with some runtime data associated with the cluster.
    data = resource.Body('data', type=dict)
    #: A list IDs of nodes that are members of the cluster.
    node_ids = resource.Body('nodes')
    #: Name of the profile used by the cluster.
    profile_name = resource.Body('profile_name')
    #: Specify whether the cluster update should only pertain to the profile.
    is_profile_only = resource.Body('profile_only', type=bool)
    #: A dictionary with dependency information of the cluster
    dependents = resource.Body('dependents', type=dict)

    def action(
        self, session: adapter.Adapter, body: dict[str, Any]
    ) -> dict[str, Any]:
        url = utils.urljoin(self.base_path, self._get_id(self), 'actions')
        resp = session.post(url, json=body)
        return cast(dict[str, Any], resp.json())

    def add_nodes(
        self, session: adapter.Adapter, nodes: list[str]
    ) -> dict[str, Any]:
        body = {
            'add_nodes': {
                'nodes': nodes,
            }
        }
        return self.action(session, body)

    def del_nodes(
        self, session: adapter.Adapter, nodes: list[str], **params: Any
    ) -> dict[str, Any]:
        data: dict[str, Any] = {'nodes': nodes}
        data.update(params)
        body = {'del_nodes': data}
        return self.action(session, body)

    def replace_nodes(
        self, session: adapter.Adapter, nodes: dict[str, str]
    ) -> dict[str, Any]:
        body = {
            'replace_nodes': {
                'nodes': nodes,
            }
        }
        return self.action(session, body)

    def scale_out(
        self, session: adapter.Adapter, count: int | None = None
    ) -> dict[str, Any]:
        body = {
            'scale_out': {
                'count': count,
            }
        }
        return self.action(session, body)

    def scale_in(
        self, session: adapter.Adapter, count: int | None = None
    ) -> dict[str, Any]:
        body = {
            'scale_in': {
                'count': count,
            }
        }
        return self.action(session, body)

    def resize(
        self, session: adapter.Adapter, **params: Any
    ) -> dict[str, Any]:
        body = {'resize': params}
        return self.action(session, body)

    def policy_attach(
        self, session: adapter.Adapter, policy_id: str, **params: Any
    ) -> dict[str, Any]:
        data: dict[str, Any] = {'policy_id': policy_id}
        data.update(params)
        body = {'policy_attach': data}
        return self.action(session, body)

    def policy_detach(
        self, session: adapter.Adapter, policy_id: str
    ) -> dict[str, Any]:
        body = {
            'policy_detach': {
                'policy_id': policy_id,
            }
        }
        return self.action(session, body)

    def policy_update(
        self, session: adapter.Adapter, policy_id: str, **params: Any
    ) -> dict[str, Any]:
        data: dict[str, Any] = {'policy_id': policy_id}
        data.update(params)
        body = {'policy_update': data}
        return self.action(session, body)

    def check(self, session: adapter.Adapter, **params: Any) -> dict[str, Any]:
        body = {'check': params}
        return self.action(session, body)

    def recover(
        self, session: adapter.Adapter, **params: Any
    ) -> dict[str, Any]:
        body = {'recover': params}
        return self.action(session, body)

    def op(
        self, session: adapter.Adapter, operation: str, **params: Any
    ) -> dict[str, Any]:
        """Perform an operation on the cluster.

        :param session: A session object used for sending request.
        :param operation: A string representing the operation to be performed.
        :param dict params: An optional dict providing the parameters for the
                            operation.
        :returns: A dictionary containing the action ID.
        """
        url = utils.urljoin(self.base_path, self.id, 'ops')
        resp = session.post(url, json={operation: params})
        return cast(dict[str, Any], resp.json())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from openstack import exceptions
from openstack.network.v2 import trunk
from openstack.tests.unit import base

EXAMPLE = {
    'id': 'IDENTIFIER',
    'description': 'Trunk description',
    'name': 'trunk-name',
    'project_id': '2',
    'admin_state_up': True,
    'port_id': 'fake_port_id',
    'status': 'ACTIVE',
    'sub_ports': [
        {
            'port_id': 'subport_port_id',
            'segmentation_id': 1234,
            'segmentation_type': 'vlan',
        }
    ],
}


class TestTrunk(base.TestCase):
    def test_basic(self):
        sot = trunk.Trunk()
        self.assertEqual('trunk', sot.resource_key)
        self.assertEqual('trunks', sot.resources_key)
        self.assertEqual('/trunks', sot.base_path)
        self.assertTrue(sot.allow_create)
        self.assertTrue(sot.allow_fetch)
        self.assertTrue(sot.allow_commit)
        self.assertTrue(sot.allow_delete)
        self.assertTrue(sot.allow_list)

    def test_make_it(self):
        sot = trunk.Trunk(**EXAMPLE)
        self.assertEqual(EXAMPLE['id'], sot.id)
        self.assertEqual(EXAMPLE['description'], sot.description)
        self.assertEqual(EXAMPLE['name'], sot.name)
        self.assertEqual(EXAMPLE['project_id'], sot.project_id)
        self.assertEqual(EXAMPLE['admin_state_up'], sot.is_admin_state_up)
        self.assertEqual(EXAMPLE['port_id'], sot.port_id)
        self.assertEqual(EXAMPLE['status'], sot.status)
        self.assertEqual(EXAMPLE['sub_ports'], sot.sub_ports)

    def test_add_subports_4xx(self):
        # Neutron may return 4xx for example if a port does not exist
        sot = trunk.Trunk(**EXAMPLE)
        response = mock.Mock()
        msg = '.*borked'
        response.body = {'NeutronError': {'message': msg}}
        response.json = mock.Mock(return_value=response.body)
        response.ok = False
        response.status_code = 404
        response.headers = {'content-type': 'application/json'}
        sess = mock.Mock()
        sess.put = mock.Mock(return_value=response)
        subports = [
            {
                'port_id': 'abc',
                'segmentation_id': '123',
                'segmentation_type': 'vlan',
            }
        ]
        with self.assertRaises(exceptions.NotFoundException, msg=msg):
            sot.add_subports(sess, subports)

    def test_delete_subports_4xx(self):
        # Neutron may return 4xx for example if a port does not exist
        sot = trunk.Trunk(**EXAMPLE)
        response = mock.Mock()
        msg = '.*borked'
        response.body = {'NeutronError': {'message': msg}}
        response.json = mock.Mock(return_value=response.body)
        response.ok = False
        response.status_code = 404
        response.headers = {'content-type': 'application/json'}
        sess = mock.Mock()
        sess.put = mock.Mock(return_value=response)
        subports = [
            {
                'port_id': 'abc',
                'segmentation_id': '123',
                'segmentation_type': 'vlan',
            }
        ]
        with self.assertRaises(exceptions.NotFoundException, msg=msg):
            sot.delete_subports(sess, subports)



# Copyright 2018 Red Hat, Inc.
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Generic, Self, TYPE_CHECKING, cast, overload
import warnings

import os_service_types

from openstack import _log
from openstack import exceptions
from openstack import proxy as proxy_mod
from openstack import warnings as os_warnings

__all__ = [
    'ServiceDescription',
]

if TYPE_CHECKING:
    from openstack.config import cloud_region
    from openstack import connection

_logger = _log.setup_logging('openstack')
_service_type_manager = os_service_types.ServiceTypes()


class _ServiceDisabledProxyShim:
    def __init__(self, service_type: str, reason: str | None) -> None:
        self.service_type = service_type
        self.reason = reason

    def __getattr__(self, item: Any) -> Any:
        raise exceptions.ServiceDisabledException(
            "Service '{service_type}' is disabled because its configuration "
            "could not be loaded. {reason}".format(
                service_type=self.service_type, reason=self.reason or ''
            )
        )


class ServiceDescription(Generic[proxy_mod.ProxyT]):
    #: Dictionary of supported versions and proxy classes for that version
    supported_versions: dict[str, type[proxy_mod.Proxy]] = {}
    #: main service_type to use to find this service in the catalog
    service_type: str
    #: list of aliases this service might be registered as
    aliases: list[str] = []

    def __init__(
        self,
        service_type: str,
        supported_versions: dict[str, type[proxy_mod.Proxy]] | None = None,
        aliases: list[str] | None = None,
    ):
        """Class describing how to interact with a REST service.

        Each service in an OpenStack cloud needs to be found by looking
        for it in the catalog. Once the endpoint is found, REST calls can
        be made, but a Proxy class and some Resource objects are needed
        to provide an object interface.

        Instances of ServiceDescription can be passed to
        `openstack.connection.Connection.add_service`, or a list can be
        passed to the `openstack.connection.Connection` constructor in
        the ``extra_services`` argument.

        All three parameters can be provided at instantation time, or
        a service-specific subclass can be used that sets the attributes
        directly.

        :param service_type: service_type to look for in the keystone catalog
        :param aliases: Optional list of aliases, if there is more than one
            name that might be used to register the service in the catalog.
        """
        self.service_type = service_type or self.service_type
        self.supported_versions = (
            supported_versions or self.supported_versions or {}
        )

        self.aliases = aliases or self.aliases
        self.all_types = [service_type, *self.aliases]

    @overload
    def __get__(self, instance: None, owner: None) -> Self: ...

    # NOTE(stephenfin): We would like to type instance as
    # connection.Connection, but due to how we construct that object, we can't
    # do so yet.
    @overload
    def __get__(
        self,
        instance: Any,
        owner: type[object],
    ) -> proxy_mod.ProxyT: ...

    def __get__(
        self,
        instance: Any,
        owner: type[object] | None,
    ) -> 'Self | proxy_mod.ProxyT':
        if instance is None:
            return self

        if self.service_type in instance._proxies:
            return cast(proxy_mod.ProxyT, instance._proxies[self.service_type])

        proxy = self._make_proxy(instance)

        if isinstance(proxy, _ServiceDisabledProxyShim):
            instance._proxies[self.service_type] = proxy
            return cast(
                proxy_mod.ProxyT,
                instance._proxies[self.service_type],
            )

        # The keystone proxy has a method called get_endpoint
        # that is about managing keystone endpoints. This is
        # unfortunate.
        try:
            endpoint = proxy_mod.Proxy.get_endpoint(proxy)
        except IndexError:
            # It's best not to look to closely here. This is
            # to support old placement.
            # There was a time when it had no status entry
            # in its version discovery doc (OY) In this case,
            # no endpoints get through version discovery
            # filtering. In order to deal with that, catch
            # the IndexError thrown by keystoneauth and
            # set an endpoint_override for the user to the
            # url in the catalog and try again.
            self._set_override_from_catalog(instance.config)
            proxy = self._make_proxy(instance)
            endpoint = proxy_mod.Proxy.get_endpoint(proxy)

        if instance._strict_proxies:
            self._validate_proxy(proxy, endpoint)

        proxy._connection = instance

        instance._proxies[self.service_type] = proxy
        return cast(proxy_mod.ProxyT, instance._proxies[self.service_type])

    def _set_override_from_catalog(
        self, config: 'cloud_region.CloudRegion'
    ) -> None:
        override = config._get_endpoint_from_catalog(
            self.service_type,
            proxy_mod.Proxy,
        )
        config.set_service_value(
            'endpoint_override',
            self.service_type,
            override,
        )

    def _validate_proxy(
        self, proxy: proxy_mod.Proxy, endpoint: str | None
    ) -> None:
        exc = None
        service_url = getattr(proxy, 'skip_discovery', None)
        try:
            # Don't go too wild for e.g. swift
            if service_url is None:
                endpoint_data = proxy.get_endpoint_data()
                service_url = (
                    endpoint_data.service_url if endpoint_data else None
                )
        except Exception as e:
            exc = e
        if exc or not endpoint or not service_url:
            raise exceptions.ServiceDiscoveryException(
                "Failed to create a working proxy for service {service_type}: "
                "{message}".format(
                    service_type=self.service_type,
                    message=exc or "No valid endpoint was discoverable.",
                )
            )

    def _make_proxy(
        self,
        instance: 'connection.Connection',
    ) -> proxy_mod.ProxyT | proxy_mod.Proxy:
        """Create a Proxy for the service in question.

        :param instance: The `openstack.connection.Connection` we're working
            with.
        """
        config = instance.config

        # This is not a valid service.
        if not config.has_service(self.service_type):
            # NOTE(stephenfin): Yes, we are lying here. But that's okay: they
            # should behave identically in a typing context
            return cast(
                proxy_mod.ProxyT,
                _ServiceDisabledProxyShim(
                    self.service_type,
                    config.get_disabled_reason(self.service_type),
                ),
            )

        # This is a valid service type, but we don't know anything about it so
        # the user is explicitly just using us for a passthrough REST adapter.
        # Skip all the lower logic.
        if not self.supported_versions:
            temp_client = config.get_session_client(
                self.service_type,
                allow_version_hack=True,
            )
            return temp_client

        # Check to see if we've got config that matches what we understand in
        # the SDK.
        version_string = config.get_api_version(self.service_type)
        endpoint_override = config.get_endpoint(self.service_type)

        # If the user doesn't give a version in config, but we only support
        # one version, then just use that version.
        if not version_string and len(self.supported_versions) == 1:
            version_string = next(iter(self.supported_versions))

        proxy_obj = None
        if endpoint_override and version_string:
            # Both endpoint override and version_string are set. We therefore
            # don't need to do discovery: just trust the user.
            proxy_class = self.supported_versions.get(version_string[0])
            if proxy_class:
                proxy_obj = config.get_session_client(
                    self.service_type,
                    constructor=proxy_class,
                )
            else:
                warnings.warn(
                    f"The configured version, {version_string} for service "
                    f"{self.service_type} is not known or supported by "
                    f"openstacksdk. The resulting Proxy object will only "
                    f"have direct passthrough REST capabilities.",
                    category=os_warnings.UnsupportedServiceVersion,
                )
        elif endpoint_override:
            temp_adapter = config.get_session_client(self.service_type)
            endpoint_data = temp_adapter.get_endpoint_data()
            if not endpoint_data:
                raise exceptions.ServiceDiscoveryException(
                    f"Failed to create a working proxy for service "
                    f"{self.service_type}: No endpoint data found."
                )

            api_version = endpoint_data.api_version
            if not api_version:
                raise exceptions.ServiceDiscoveryException(
                    f"Failed to create a working proxy for service "
                    f"{self.service_type}: No version in endpoint data."
                )

            proxy_class = self.supported_versions.get(str(api_version[0]))
            if proxy_class:
                proxy_obj = config.get_session_client(
                    self.service_type,
                    constructor=proxy_class,
                )
            else:
                warnings.warn(
                    f"Service {self.service_type} has an endpoint override "
                    f"set but the version discovered at that endpoint, "
                    f"{api_version}, is not supported by openstacksdk. "
                    f"The resulting Proxy object will only have direct "
                    f"passthrough REST capabilities.",
                    category=os_warnings.UnsupportedServiceVersion,
                )

        if proxy_obj:
            if getattr(proxy_obj, 'skip_discovery', False):
                # Some services, like swift, don't have discovery. While
                # keystoneauth will behave correctly and handle such
                # scenarios, it's not super efficient as it involves trying
                # and falling back a few times.
                return proxy_obj

            data = proxy_obj.get_endpoint_data()
            if not data:
                if instance._strict_proxies:
                    raise exceptions.ServiceDiscoveryException(
                        f"Failed to create a working proxy for service "
                        f"{self.service_type}: No endpoint data found."
                    )
                else:
                    return proxy_obj

            # If we've gotten here with a proxy object it means we have
            # an endpoint_override in place. If the catalog_url and
            # service_url don't match, which can happen if there is a
            # None plugin and auth.endpoint like with standalone ironic,
            # we need to be explicit that this service has an endpoint_override
            # so that subsequent discovery calls don't get made incorrectly.
            if data.catalog_url != data.service_url:
                ep_key = '{service_type}_endpoint_override'.format(
                    service_type=self.service_type.replace('-', '_')
                )
                config.config[ep_key] = data.service_url
                proxy_obj = config.get_session_client(
                    self.service_type,
                    constructor=proxy_class,
                )

            return proxy_obj

        # Make an adapter to let discovery take over
        supported_versions = sorted([int(f) for f in self.supported_versions])
        if version_string:
            if getattr(
                self.supported_versions[str(supported_versions[0])],
                'skip_discovery',
                False,
            ):
                # Requested service does not support version discovery
                # In this case it is more efficient to set the
                # endpoint_override to the current catalog endpoint value,
                # otherwise next request will try to perform discovery.

                temp_adapter = config.get_session_client(self.service_type)
                ep_override = temp_adapter.get_endpoint(skip_discovery=True)

                ep_key = '{service_type}_endpoint_override'.format(
                    service_type=self.service_type.replace('-', '_')
                )
                config.config[ep_key] = ep_override

                return config.get_session_client(
                    self.service_type,
                    version=version_string,
                    constructor=self.supported_versions[
                        str(supported_versions[0])
                    ],
                    allow_version_hack=True,
                )

            temp_adapter = config.get_session_client(
                self.service_type,
                allow_version_hack=True,
                version=version_string,
            )
        else:
            temp_adapter = config.get_session_client(
                self.service_type,
                allow_version_hack=True,
                max_version=f'{supported_versions[-1]!s}.latest',
                min_version=f'{supported_versions[0]!s}',
            )

        found_version = temp_adapter.get_api_major_version()
        if found_version is None:
            region_name = instance.config.get_region_name(self.service_type)
            raise exceptions.NotSupported(
                f"The {self.service_type} service for "
                f"{instance.name}:{region_name} exists but does not have "
                f"any supported versions."
            )

        proxy_class = self.supported_versions.get(str(found_version[0]))
        if proxy_class:
            if version_string:
                return config.get_session_client(
                    self.service_type,
                    constructor=proxy_class,
                    allow_version_hack=True,
                    version=version_string,
                )

            return config.get_session_client(
                self.service_type,
                version=version_string,
                constructor=proxy_class,
                allow_version_hack=True,
                max_version=f'{supported_versions[-1]!s}.latest',
                min_version=f'{supported_versions[0]!s}',
            )

        # No proxy_class
        # Maybe openstacksdk is being used for the passthrough
        # REST API proxy layer for an unknown service in the
        # service catalog that also doesn't have any useful
        # version discovery?
        warnings.warn(
            f"Service {self.service_type} has no discoverable version. "
            "The resulting Proxy object will only have direct "
            "passthrough REST capabilities.",
            category=os_warnings.UnsupportedServiceVersion,
        )
        return temp_adapter

    def __set__(self, instance: Any, value: Any) -> None:
        raise AttributeError('Service Descriptors cannot be set')

    def __delete__(self, instance: Any) -> None:
        # NOTE(gtema) Some clouds are not very fast (or interested at all)
        # in bringing their changes upstream. If there are incompatible changes
        # downstream we need to allow overriding default implementation by
        # deleting service_type attribute of the connection and then
        # "add_service" with new implementation.
        # This is intentionally designed to be hard to use to show how bad it
        # is not to contribute changes back
        for service_type in self.all_types:
            if service_type in instance._proxies:
                del instance._proxies[service_type]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
from unittest import mock

import fixtures
from keystoneauth1 import identity
from keystoneauth1 import session

import openstack.config
from openstack import connection
from openstack import proxy
from openstack import service_description
from openstack.tests.unit import base
from openstack.tests.unit.fake import fake_service


CONFIG_AUTH_URL = "https://identity.example.com/"
CONFIG_USERNAME = "BozoTheClown"
CONFIG_PASSWORD = "TopSecret"
CONFIG_PROJECT = "TheGrandPrizeGame"
CONFIG_CACERT = "TrustMe"

CLOUD_CONFIG = f"""
clouds:
  sample-cloud:
    region_name: RegionOne
    auth:
      auth_url: {CONFIG_AUTH_URL}
      username: {CONFIG_USERNAME}
      password: {CONFIG_PASSWORD}
      project_name: {CONFIG_PROJECT}
  insecure-cloud:
    auth:
      auth_url: {CONFIG_AUTH_URL}
      username: {CONFIG_USERNAME}
      password: {CONFIG_PASSWORD}
      project_name: {CONFIG_PROJECT}
    cacert: {CONFIG_CACERT}
    verify: False
  insecure-cloud-alternative-format:
    auth:
      auth_url: {CONFIG_AUTH_URL}
      username: {CONFIG_USERNAME}
      password: {CONFIG_PASSWORD}
      project_name: {CONFIG_PROJECT}
    insecure: True
  cacert-cloud:
    auth:
      auth_url: {CONFIG_AUTH_URL}
      username: {CONFIG_USERNAME}
      password: {CONFIG_PASSWORD}
      project_name: {CONFIG_PROJECT}
    cacert: {CONFIG_CACERT}
  profiled-cloud:
    profile: dummy
    auth:
      username: {CONFIG_USERNAME}
      password: {CONFIG_PASSWORD}
      project_name: {CONFIG_PROJECT}
    cacert: {CONFIG_CACERT}
"""

VENDOR_CONFIG = f"""
{{
  "name": "dummy",
  "profile": {{
    "auth": {{
      "auth_url": "{CONFIG_AUTH_URL}"
    }},
    "vendor_hook": "openstack.tests.unit.test_connection:vendor_hook"
  }}
}}
"""

PUBLIC_CLOUDS_YAML = f"""
public-clouds:
  dummy:
    auth:
      auth_url: {CONFIG_AUTH_URL}
    vendor_hook: openstack.tests.unit.test_connection:vendor_hook
"""


class _TestConnectionBase(base.TestCase):
    def setUp(self):
        super().setUp()
        # Create a temporary directory where our test config will live
        # and insert it into the search path via OS_CLIENT_CONFIG_FILE.
        config_dir = self.useFixture(fixtures.TempDir()).path
        config_path = os.path.join(config_dir, "clouds.yaml")

        with open(config_path, "w") as conf:
            conf.write(CLOUD_CONFIG)

        self.useFixture(
            fixtures.EnvironmentVariable("OS_CLIENT_CONFIG_FILE", config_path)
        )
        self.use_keystone_v2()


class TestConnection(_TestConnectionBase):
    def test_other_parameters(self):
        conn = connection.Connection(cloud='sample-cloud', cert='cert')
        self.assertEqual(conn.session.cert, 'cert')

    def test_session_provided(self):
        mock_session = mock.Mock(spec=session.Session)
        mock_session.auth = identity.V3Password(
            auth_url='https://auth.example.com',
            password='passw0rd',
            user_id='fake',
        )
        conn = connection.Connection(session=mock_session, cert='cert')
        self.assertEqual(mock_session, conn.session)
        self.assertEqual('auth.example.com', conn.config.name)

    def test_create_session(self):
        conn = connection.Connection(cloud='sample-cloud')
        self.assertIsNotNone(conn)
        # TODO(mordred) Rework this - we need to provide requests-mock
        # entries for each of the proxies below
        # self.assertEqual('openstack.proxy',
        #                  conn.alarm.__class__.__module__)
        # self.assertEqual('openstack.clustering.v1._proxy',
        #                  conn.clustering.__class__.__module__)
        # self.assertEqual('openstack.compute.v2._proxy',
        #                  conn.compute.__class__.__module__)
        # self.assertEqual('openstack.database.v1._proxy',
        #                  conn.database.__class__.__module__)
        # self.assertEqual('openstack.identity.v2._proxy',
        #                  conn.identity.__class__.__module__)
        # self.assertEqual('openstack.image.v2._proxy',
        #                  conn.image.__class__.__module__)
        # self.assertEqual('openstack.object_store.v1._proxy',
        #                  conn.object_store.__class__.__module__)
        # self.assertEqual('openstack.load_balancer.v2._proxy',
        #                  conn.load_balancer.__class__.__module__)
        # self.assertEqual('openstack.orchestration.v1._proxy',
        #                  conn.orchestration.__class__.__module__)
        # self.assertEqual('openstack.workflow.v2._proxy',
        #                  conn.workflow.__class__.__module__)

    def test_create_unknown_proxy(self):
        self.register_uris(
            [
                self.get_placement_discovery_mock_dict(),
            ]
        )

        def closure():
            return self.cloud.placement

        self.assertIsInstance(self.cloud.placement, proxy.Proxy)
        self.assert_calls()

    def test_create_connection_version_param_default(self):
        c1 = connection.Connection(cloud='sample-cloud')
        conn = connection.Connection(session=c1.session)
        self.assertEqual(
            'openstack.identity.v3._proxy', conn.identity.__class__.__module__
        )

    def test_create_connection_version_param_string(self):
        c1 = connection.Connection(cloud='sample-cloud')
        conn = connection.Connection(
            session=c1.session, identity_api_version='2'
        )
        self.assertEqual(
            'openstack.identity.v2._proxy', conn.identity.__class__.__module__
        )

    def test_create_connection_version_param_int(self):
        c1 = connection.Connection(cloud='sample-cloud')
        conn = connection.Connection(
            session=c1.session, identity_api_version=3
        )
        self.assertEqual(
            'openstack.identity.v3._proxy', conn.identity.__class__.__module__
        )

    def test_create_connection_version_param_bogus(self):
        c1 = connection.Connection(cloud='sample-cloud')
        conn = connection.Connection(
            session=c1.session, identity_api_version='red'
        )
        # TODO(mordred) This is obviously silly behavior
        self.assertEqual(
            'openstack.identity.v3._proxy', conn.identity.__class__.__module__
        )

    def test_from_config_given_config(self):
        cloud_region = openstack.config.OpenStackConfig().get_one(
            "sample-cloud"
        )

        sot = connection.from_config(config=cloud_region)

        self.assertEqual(
            CONFIG_USERNAME, sot.config.config['auth']['username']
        )
        self.assertEqual(
            CONFIG_PASSWORD, sot.config.config['auth']['password']
        )
        self.assertEqual(
            CONFIG_AUTH_URL, sot.config.config['auth']['auth_url']
        )
        self.assertEqual(
            CONFIG_PROJECT, sot.config.config['auth']['project_name']
        )

    def test_from_config_given_cloud(self):
        sot = connection.from_config(cloud="sample-cloud")

        self.assertEqual(
            CONFIG_USERNAME, sot.config.config['auth']['username']
        )
        self.assertEqual(
            CONFIG_PASSWORD, sot.config.config['auth']['password']
        )
        self.assertEqual(
            CONFIG_AUTH_URL, sot.config.config['auth']['auth_url']
        )
        self.assertEqual(
            CONFIG_PROJECT, sot.config.config['auth']['project_name']
        )

    def test_from_config_given_cloud_config(self):
        cloud_region = openstack.config.OpenStackConfig().get_one(
            "sample-cloud"
        )

        sot = connection.from_config(cloud_config=cloud_region)

        self.assertEqual(
            CONFIG_USERNAME, sot.config.config['auth']['username']
        )
        self.assertEqual(
            CONFIG_PASSWORD, sot.config.config['auth']['password']
        )
        self.assertEqual(
            CONFIG_AUTH_URL, sot.config.config['auth']['auth_url']
        )
        self.assertEqual(
            CONFIG_PROJECT, sot.config.config['auth']['project_name']
        )

    def test_from_config_given_cloud_name(self):
        sot = connection.from_config(cloud_name="sample-cloud")

        self.assertEqual(
            CONFIG_USERNAME, sot.config.config['auth']['username']
        )
        self.assertEqual(
            CONFIG_PASSWORD, sot.config.config['auth']['password']
        )
        self.assertEqual(
            CONFIG_AUTH_URL, sot.config.config['auth']['auth_url']
        )
        self.assertEqual(
            CONFIG_PROJECT, sot.config.config['auth']['project_name']
        )

    def test_from_config_verify(self):
        sot = connection.from_config(cloud="insecure-cloud")
        self.assertFalse(sot.session.verify)

        sot = connection.from_config(cloud="cacert-cloud")
        self.assertEqual(CONFIG_CACERT, sot.session.verify)

    def test_from_config_insecure(self):
        # Ensure that the "insecure=True" flag implies "verify=False"
        sot = connection.from_config("insecure-cloud-alternative-format")
        self.assertFalse(sot.session.verify)


class TestOsloConfig(_TestConnectionBase):
    def test_from_conf(self):
        c1 = connection.Connection(cloud='sample-cloud')
        conn = connection.Connection(
            session=c1.session, oslo_conf=self._load_ks_cfg_opts()
        )
        # There was no config for keystone
        self.assertIsInstance(
            conn.identity, service_description._ServiceDisabledProxyShim
        )
        # But nova was in there
        self.assertEqual(
            'openstack.compute.v2._proxy', conn.compute.__class__.__module__
        )

    def test_from_conf_filter_service_types(self):
        c1 = connection.Connection(cloud='sample-cloud')
        conn = connection.Connection(
            session=c1.session,
            oslo_conf=self._load_ks_cfg_opts(),
            service_types={'orchestration', 'i-am-ignored'},
        )
        # There was no config for keystone
        self.assertIsInstance(
            conn.identity, service_description._ServiceDisabledProxyShim
        )
        # Nova was in there, but disabled because not requested
        self.assertIsInstance(
            conn.compute, service_description._ServiceDisabledProxyShim
        )


class TestNetworkConnection(base.TestCase):
    # Verify that if the catalog has the suffix we don't mess things up.
    def test_network_proxy(self):
        self.os_fixture.v3_token.remove_service('network')
        svc = self.os_fixture.v3_token.add_service('network')
        svc.add_endpoint(
            interface='public',
            url='https://network.example.com/v2.0',
            region='RegionOne',
        )
        self.use_keystone_v3()
        self.assertEqual(
            'openstack.network.v2._proxy',
            self.cloud.network.__class__.__module__,
        )
        self.assert_calls()
        self.assertEqual(
            "https://network.example.com/v2.0",
            self.cloud.network.get_endpoint(),
        )


class TestNetworkConnectionSuffix(base.TestCase):
    # We need to do the neutron adapter test differently because it needs
    # to actually get a catalog.

    def test_network_proxy(self):
        self.assertEqual(
            'openstack.network.v2._proxy',
            self.cloud.network.__class__.__module__,
        )
        self.assert_calls()
        self.assertEqual(
            "https://network.example.com/v2.0",
            self.cloud.network.get_endpoint(),
        )


class TestAuthorize(base.TestCase):
    def test_authorize_works(self):
        res = self.cloud.authorize()
        self.assertEqual('KeystoneToken-1', res)

    def test_authorize_failure(self):
        self.use_broken_keystone()

        self.assertRaises(
            openstack.exceptions.SDKException, self.cloud.authorize
        )


class TestNewService(base.TestCase):
    def test_add_service_v1(self):
        svc = self.os_fixture.v3_token.add_service('fake')
        svc.add_endpoint(
            interface='public',
            region='RegionOne',
            url=f'https://fake.example.com/v1/{base.PROJECT_ID}',
        )
        self.use_keystone_v3()
        conn = self.cloud

        service = fake_service.FakeService('fake')

        conn.add_service(service)

        # Ensure no discovery calls made
        self.assertEqual(0, len(self.adapter.request_history))

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri='https://fake.example.com',
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri='https://fake.example.com/v1/',
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url('fake'),
                    status_code=404,
                ),
            ]
        )

        self.assertEqual(
            'openstack.tests.unit.fake.v1._proxy',
            conn.fake.__class__.__module__,
        )
        self.assertTrue(conn.fake.dummy())

    def test_add_service_v2(self):
        svc = self.os_fixture.v3_token.add_service('fake')
        svc.add_endpoint(
            interface='public',
            region='RegionOne',
            url=f'https://fake.example.com/v2/{base.PROJECT_ID}',
        )
        self.use_keystone_v3()
        conn = self.cloud

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri='https://fake.example.com',
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri='https://fake.example.com/v2/',
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url('fake'),
                    status_code=404,
                ),
            ]
        )

        service = fake_service.FakeService('fake')

        conn.add_service(service)

        self.assertEqual(
            'openstack.tests.unit.fake.v2._proxy',
            conn.fake.__class__.__module__,
        )
        self.assertFalse(conn.fake.dummy())

    def test_replace_system_service(self):
        svc = self.os_fixture.v3_token.add_service('fake')
        svc.add_endpoint(
            interface='public',
            region='RegionOne',
            url=f'https://fake.example.com/v2/{base.PROJECT_ID}',
        )
        self.use_keystone_v3()
        conn = self.cloud

        # delete native dns service
        delattr(conn, 'dns')

        self.register_uris(
            [
                dict(
                    method='GET',
                    uri='https://fake.example.com',
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri='https://fake.example.com/v2/',
                    status_code=404,
                ),
                dict(
                    method='GET',
                    uri=self.get_mock_url('fake'),
                    status_code=404,
                ),
            ]
        )

        # add fake service with alias 'DNS'
        service = fake_service.FakeService('fake', aliases=['dns'])
        conn.add_service(service)

        # ensure dns service responds as we expect from replacement
        self.assertFalse(conn.dns.dummy())


def vendor_hook(conn):
    setattr(conn, 'test', 'test_val')


class TestVendorProfile(base.TestCase):
    def setUp(self):
        super().setUp()
        # Create a temporary directory where our test config will live
        # and insert it into the search path via OS_CLIENT_CONFIG_FILE.
        config_dir = self.useFixture(fixtures.TempDir()).path
        config_path = os.path.join(config_dir, "clouds.yaml")
        public_clouds = os.path.join(config_dir, "clouds-public.yaml")

        with open(config_path, "w") as conf:
            conf.write(CLOUD_CONFIG)

        with open(public_clouds, "w") as conf:
            conf.write(PUBLIC_CLOUDS_YAML)

        self.useFixture(
            fixtures.EnvironmentVariable("OS_CLIENT_CONFIG_FILE", config_path)
        )
        self.use_keystone_v2()

        self.config = openstack.config.loader.OpenStackConfig(
            vendor_files=[public_clouds]
        )

    def test_conn_from_profile(self):
        self.cloud = self.config.get_one(cloud='profiled-cloud')

        conn = connection.Connection(config=self.cloud)

        self.assertIsNotNone(conn)

    def test_hook_from_profile(self):
        self.cloud = self.config.get_one(cloud='profiled-cloud')

        conn = connection.Connection(config=self.cloud)

        self.assertEqual('test_val', conn.test)

    def test_hook_from_connection_param(self):
        conn = connection.Connection(
            cloud='sample-cloud',
            vendor_hook='openstack.tests.unit.test_connection:vendor_hook',
        )

        self.assertEqual('test_val', conn.test)

    def test_hook_from_connection_ignore_missing(self):
        conn = connection.Connection(
            cloud='sample-cloud',
            vendor_hook='openstack.tests.unit.test_connection:missing',
        )

        self.assertIsNotNone(conn)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.tests.functional import base


class TestServiceStatus(base.BaseFunctionalTest):
    def setUp(self):
        super().setUp()
        self.require_service('dns')

        self.service_names = [
            "api",
            "backend",
            "central",
            "mdns",
            "producer",
            "sink",
            "storage",
            "worker",
        ]
        self.service_status = ["UP", "DOWN"]

    def test_service_status(self):
        service_statuses = list(self.operator_cloud.dns.service_statuses())
        if not service_statuses:
            self.skipTest(
                "The Service in Designate System is required for this test"
            )

        names = [f.service_name for f in service_statuses]
        statuses = [f.status for f in service_statuses]

        self.assertTrue(
            all(status in self.service_status for status in statuses)
        )
        self.assertTrue(all(name in self.service_names for name in names))

        # Test that we can fetch a service status
        service_status = self.operator_cloud.dns.get_service_status(
            service_statuses[0]
        )
        self.assertIn(service_status.service_name, self.service_names)
        self.assertIn(service_status.status, self.service_status)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Self

from keystoneauth1 import adapter

from openstack import resource
from openstack import utils


class Cluster(resource.Resource):
    resource_key = 'cluster'
    resources_key = 'clusters'
    base_path = '/clusters'

    # capabilities
    allow_fetch = True
    allow_list = True

    _max_microversion = '3.26'

    _query_mapping = resource.QueryParameters(
        'name',
        'binary',
        'is_up',
        'disabled',
        'num_hosts',
        'num_down_hosts',
        'replication_status',
        'frozen',
        'active_backend_id',
    )

    # Properties
    #: The ID of active storage backend (cinder-volume services only).
    active_backend_id = resource.Body('active_backend_id')
    #: The binary name of the services in the cluster.
    binary = resource.Body('binary')
    #: The date and time when the resource was created.
    created_at = resource.Body('created_at')
    #: The reason for disabling the cluster.
    disabled_reason = resource.Body('disabled_reason')
    #: Whether the cluster is frozen for replication.
    frozen = resource.Body('frozen', type=bool)
    #: The last periodic heartbeat received.
    last_heartbeat = resource.Body('last_heartbeat')
    #: The name of the cluster.
    name = resource.Body('name', alternate_id=True)
    #: The number of down hosts in the cluster.
    num_down_hosts = resource.Body('num_down_hosts', type=int)
    #: The number of hosts in the cluster.
    num_hosts = resource.Body('num_hosts', type=int)
    #: The cluster replication status.
    replication_status = resource.Body('replication_status')
    #: The state of the cluster. One of ``up`` or ``down``.
    state = resource.Body('state')
    #: The status of the cluster. One of ``enabled`` or ``disabled``.
    status = resource.Body('status')
    #: The date and time when the resource was last updated.
    updated_at = resource.Body('updated_at')

    def _action(
        self,
        session: adapter.Adapter,
        action: str,
        body: dict[str, Any],
    ) -> Self:
        url = utils.urljoin(Cluster.base_path, action)
        microversion = session.default_microversion
        response = session.put(url, json=body, microversion=microversion)
        self._translate_response(response)
        return self

    def enable(self, session: adapter.Adapter) -> Self:
        """Enable scheduling for the cluster."""
        body: dict[str, Any] = {'name': self.name}
        if self.binary:
            body['binary'] = self.binary
        return self._action(session, 'enable', body)

    def disable(
        self,
        session: adapter.Adapter,
        *,
        reason: str | None = None,
    ) -> Self:
        """Disable scheduling for the cluster.

        :param reason: The reason for disabling the cluster.
        """
        body: dict[str, Any] = {'name': self.name}
        if self.binary:
            body['binary'] = self.binary
        if reason:
            body['disabled_reason'] = reason
        return self._action(session, 'disable', body)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Self

from keystoneauth1 import adapter
import requests

from openstack.common import metadata
from openstack import exceptions
from openstack import resource
from openstack import utils


class ShareSnapshot(resource.Resource, metadata.MetadataMixin):
    resource_key = "snapshot"
    resources_key = "snapshots"
    base_path = "/snapshots"

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True
    allow_head = False

    _query_mapping = resource.QueryParameters("snapshot_id")

    #: Properties
    #: The date and time stamp when the resource was
    #: created within the services's database.
    created_at = resource.Body("created_at")
    #: The user defined description of the resource.
    description = resource.Body("description", type=str)
    #: The user defined name of the resource.
    display_name = resource.Body("display_name", type=str)
    #: The user defined description of the resource
    display_description = resource.Body("display_description", type=str)
    #: ID of the project that the snapshot belongs to.
    project_id = resource.Body("project_id", type=str)
    #: The UUID of the source share that was used to
    #: create the snapshot.
    share_id = resource.Body("share_id", type=str)
    #: The file system protocol of a share snapshot
    share_proto = resource.Body("share_proto", type=str)
    #: The snapshot's source share's size, in GiBs.
    share_size = resource.Body("share_size", type=int)
    #: The snapshot size, in GiBs.
    size = resource.Body("size", type=int)
    #: The snapshot status
    status = resource.Body("status", type=str)
    #: ID of the user that the snapshot was created by.
    user_id = resource.Body("user_id", type=str)

    def _action(
        self,
        session: adapter.Adapter,
        body: dict[str, Any],
        microversion: str | None = None,
    ) -> requests.Response:
        url = utils.urljoin(self.base_path, self.id, 'action')
        headers = {'Accept': ''}

        if microversion is None:
            microversion = self._get_microversion(session)

        response = session.post(
            url, json=body, headers=headers, microversion=microversion
        )

        exceptions.raise_from_response(response)
        return response

    def reset_status(self, session: adapter.Adapter, status: str) -> None:
        """Reset the snapshot to the given status.

        :param status: The status of the share to reset to.
        :returns: ``None``
        """
        body = {'reset_status': {'status': status}}
        self._action(session, body)

    def force_delete(self, session: adapter.Adapter) -> None:
        """Force delete the snapshot.

        :returns: ``None``
        """
        body = {'force_delete': None}
        self._action(session, body)

    def manage(
        self,
        session: adapter.Adapter,
        share_id: str,
        provider_location: str,
        **params: Any,
    ) -> Self:
        """Manage a share snapshot.

        :param session: A session object used for sending request.
        :param share_id: The UUID of the share that has snapshot which
            should be managed.
        :param provider_location: Provider location of the snapshot on the
            backend.
        :param params: Optional parameters to be sent. Available
            parameters include:

            * name: The user defined name of the resource.
            * display_name: The user defined name of the resource. This field
              sets the name parameter.
            * description: The user defined description of the resource.
            * display_description: The user defined description of the
              resource. This field sets the description parameter.
            * driver_options: A set of one or more key and value pairs, as a
              dictionary of strings, that describe driver options.

        :returns: The share snapshot that was managed.
        """
        attrs = {
            'snapshot': {
                'share_id': share_id,
                'provider_location': provider_location,
            }
        }
        attrs['snapshot'].update(params)

        url = utils.urljoin(self.base_path, 'manage')
        resp = session.post(url, json=attrs)
        self._translate_response(resp)
        return self

    def unmanage(self, session: adapter.Adapter) -> None:
        """Unmanage a share snapshot.

        :param session: A session object used for sending request.
        :returns: ``None``
        """
        body = {'unmanage': None}
        self._action(session, body)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import warnings
from typing import Any, Self

import requests

from keystoneauth1 import adapter

from openstack import exceptions
from openstack import resource
from openstack import utils
from openstack import warnings as os_warnings


class Group(resource.Resource):
    resource_key = "group"
    resources_key = "groups"
    base_path = "/groups"

    # capabilities
    allow_fetch = True
    allow_create = True
    allow_delete = True
    allow_commit = True
    allow_list = True

    _query_mapping = resource.QueryParameters(
        "limit",
        "marker",
        "offset",
        "sort_dir",
        "sort_key",
        "sort",
        all_projects="all_tenants",
    )

    availability_zone = resource.Body("availability_zone")
    created_at = resource.Body("created_at")
    description = resource.Body("description")
    group_snapshot_id = resource.Body("group_snapshot_id")
    group_type = resource.Body("group_type")
    project_id = resource.Body("project_id")
    replication_targets = resource.Body("replication_targets", type=list)
    replication_status = resource.Body("replication_status")
    source_group_id = resource.Body("source_group_id")
    status = resource.Body("status")
    volumes = resource.Body("volumes", type=list)
    volume_types = resource.Body("volume_types", type=list)

    _max_microversion = "3.38"

    def _action(
        self, session: adapter.Adapter, body: dict[str, Any]
    ) -> requests.Response:
        """Preform group actions given the message body."""
        session = self._get_session(session)
        microversion = self._get_microversion(session)
        url = utils.urljoin(self.base_path, self.id, 'action')
        response = session.post(url, json=body, microversion=microversion)
        exceptions.raise_from_response(response)
        return response

    def delete(
        self,
        session: adapter.Adapter,
        error_message: str | None = None,
        *,
        microversion: str | None = None,
        delete_volumes: bool = False,
        **attrs: Any,
    ) -> Self:
        """Delete a group."""
        body = {'delete': {'delete-volumes': delete_volumes}}
        self._action(session, body)
        return self

    def fetch_replication_targets(self, session: adapter.Adapter) -> Self:
        """Fetch replication targets for the group.

        :param session: The session to use for making this request.
        :return: This group with the ``replication_targets`` field populated.
        """
        body = {'list_replication_targets': None}
        response = self._action(session, body)
        self._body.attributes.update(
            {'replication_targets': response.json()['replication_targets']}
        )
        return self

    def enable_replication(self, session: adapter.Adapter) -> None:
        """Enable replication for the group.

        :param session: The session to use for making this request.
        """
        body = {'enable_replication': None}
        self._action(session, body)

    def disable_replication(self, session: adapter.Adapter) -> None:
        """Disable replication for the group.

        :param session: The session to use for making this request.
        """
        body = {'disable_replication': None}
        self._action(session, body)

    def failover_replication(
        self,
        session: adapter.Adapter,
        *,
        allowed_attached_volume: bool = False,
        secondary_backend_id: str | None = None,
    ) -> None:
        """Failover replication for the group.

        :param session: The session to use for making this request.
        :param allowed_attached_volume: Whether to allow attached volumes in
            the group.
        :param secondary_backend_id: The secondary backend ID.
        :returns: None
        """
        body = {
            'modify_body_for_action': {
                'allow_attached_volume': allowed_attached_volume,
                'secondary_backend_id': secondary_backend_id,
            },
        }
        self._action(session, body)

    def reset_status(self, session: adapter.Adapter, status: str) -> None:
        """Resets the status for a group.

        :param session: The session to use for making this request.
        :param status: The status for the group.
        """
        body = {'reset_status': {'status': status}}
        self._action(session, body)

    def reset(self, session: adapter.Adapter, status: str) -> None:
        warnings.warn(
            "reset is a deprecated alias for reset_status and will be "
            "removed in a future release.",
            os_warnings.RemovedInSDK60Warning,
        )
        self.reset_status(session, status)

    @classmethod
    def create_from_source(
        cls,
        session: adapter.Adapter,
        group_snapshot_id: str | None,
        source_group_id: str | None,
        name: str | None = None,
        description: str | None = None,
    ) -> Self:
        """Creates a new group from source."""
        session = cls._get_session(session)
        microversion = cls._get_microversion(session)
        url = utils.urljoin(cls.base_path, 'action')
        body = {
            'create-from-src': {
                'name': name,
                'description': description,
                'group_snapshot_id': group_snapshot_id,
                'source_group_id': source_group_id,
            }
        }
        response = session.post(url, json=body, microversion=microversion)
        exceptions.raise_from_response(response)

        group = cls()
        group._translate_response(response)
        return group



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack.block_storage.v2 import stats
from openstack.tests.unit import base


POOLS = {
    "name": "pool1",
    "capabilities": {
        "updated": "2014-10-28T00=00=00-00=00",
        "total_capacity": 1024,
        "free_capacity": 100,
        "volume_backend_name": "pool1",
        "reserved_percentage": "0",
        "driver_version": "1.0.0",
        "storage_protocol": "iSCSI",
        "QoS_support": "false",
    },
}


class TestBackendPools(base.TestCase):
    def setUp(self):
        super().setUp()

    def test_basic(self):
        sot = stats.Pools(POOLS)
        self.assertEqual("", sot.resource_key)
        self.assertEqual("pools", sot.resources_key)
        self.assertEqual(
            "/scheduler-stats/get_pools?detail=True", sot.base_path
        )
        self.assertFalse(sot.allow_create)
        self.assertFalse(sot.allow_fetch)
        self.assertFalse(sot.allow_delete)
        self.assertTrue(sot.allow_list)
        self.assertFalse(sot.allow_commit)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""
The :mod:`~openstack.test.fakes` module exists to help application developers
using the OpenStack SDK to unit test their applications. It provides a number
of helper utilities to generate fake :class:`~openstack.resource.Resource` and
:class:`~openstack.proxy.Proxy` instances. These fakes do not require an
established connection and allow you to validate that your application using
valid attributes and methods for both :class:`~openstack.resource.Resource` and
:class:`~openstack.proxy.Proxy` instances.
"""

from collections.abc import Generator
import inspect
import random
import types
import typing
from typing import Any, cast
from unittest import mock
import uuid

from openstack import fields
from openstack import format as _format
from openstack import proxy
from openstack import resource
from openstack import service_description


def generate_fake_resource(
    resource_type: type[resource.ResourceT],
    **attrs: Any,
) -> resource.ResourceT:
    """Generate a fake resource

    Example usage:

    .. code-block:: python

        >>> from openstack.compute.v2 import server
        >>> from openstack.test import fakes
        >>> fakes.generate_fake_resource(server.Server)
        openstack.compute.v2.server.Server(...)

    :param resource_type: Object class
    :param attrs: Optional attributes to be set on resource
    :return: Instance of ``resource_type`` class populated with fake
        values of expected types
    :raises NotImplementedError: If a resource attribute specifies a ``type``
        or ``list_type`` that cannot be automatically generated
    """
    base_attrs: dict[str, Any] = {}
    for name, value in inspect.getmembers(
        resource_type,
        predicate=lambda x: isinstance(x, fields.Body | fields.URI),
    ):
        if isinstance(value, fields.Body):
            target_type = value.type
            if target_type is None:
                if (
                    name == "properties"
                    and hasattr(
                        resource_type, "_store_unknown_attrs_as_properties"
                    )
                    and resource_type._store_unknown_attrs_as_properties
                ):
                    # virtual "properties" attr which hosts all unknown attrs
                    # (i.e. Image)
                    base_attrs[name] = dict()
                else:
                    # Type not defined - string
                    base_attrs[name] = uuid.uuid4().hex
            elif issubclass(target_type, resource.Resource):
                # Attribute is of another Resource type
                base_attrs[name] = generate_fake_resource(target_type)
            elif issubclass(target_type, list) and value.list_type is not None:
                # List of ...
                item_type = value.list_type
                if issubclass(item_type, resource.Resource):
                    # item is of Resource type
                    base_attrs[name] = [generate_fake_resource(item_type)]
                elif issubclass(item_type, dict):
                    base_attrs[name] = [{}]
                elif issubclass(item_type, str):
                    base_attrs[name] = [uuid.uuid4().hex]
                else:
                    # Everything else
                    msg = (
                        f"Fake value for {resource_type.__name__}.{name} can "
                        f"not be generated"
                    )
                    raise NotImplementedError(msg)
            elif issubclass(target_type, list) and value.list_type is None:
                # List of str
                base_attrs[name] = [uuid.uuid4().hex]
            elif issubclass(target_type, str):
                # definitely string
                base_attrs[name] = uuid.uuid4().hex
            elif issubclass(target_type, int):
                # int
                base_attrs[name] = random.randint(1, 100)  # noqa: S311
            elif issubclass(target_type, float):
                # float
                base_attrs[name] = random.random()  # noqa: S311
            elif issubclass(target_type, bool) or issubclass(
                target_type, _format.BoolStr
            ):
                # bool
                base_attrs[name] = random.choice([True, False])  # noqa: S311
            elif issubclass(target_type, dict):
                # some dict - without further details leave it empty
                base_attrs[name] = dict()
            else:
                # Everything else
                msg = (
                    f"Fake value for {resource_type.__name__}.{name} can not "
                    f"be generated"
                )
                raise NotImplementedError(msg)

        if isinstance(value, fields.URI):
            # For URI we just generate something
            base_attrs[name] = uuid.uuid4().hex

    base_attrs.update(**attrs)
    fake = resource_type(**base_attrs)
    return fake


def generate_fake_resources(
    resource_type: type[resource.ResourceT],
    count: int = 1,
    attrs: dict[str, Any] | None = None,
) -> Generator[resource.ResourceT, None, None]:
    """Generate a given number of fake resource entities

    Example usage:

    .. code-block:: python

        >>> from openstack.compute.v2 import server
        >>> from openstack.test import fakes
        >>> fakes.generate_fake_resources(server.Server, count=3)
        <generator object generate_fake_resources at 0x7f075dc65040>

    :param resource_type: Object class
    :param count: Number of objects to return
    :param attrs: Attribute values to set into each instance
    :return: Generator of ``resource_type`` class instances populated with fake
        values of expected types.
    """
    if not attrs:
        attrs = {}
    for _ in range(count):
        yield generate_fake_resource(resource_type, **attrs)


def _resource_type_from_hint(
    return_hint: Any,
) -> tuple[type[resource.Resource], bool] | None:
    """Extract a resource type and generator flag from a return type hint.

    Returns ``(resource_type, is_generator)`` or ``None`` if the hint does not
    describe a Resource (or Generator thereof).
    """
    # Direct Resource subclass: -> Server
    if isinstance(return_hint, type) and issubclass(
        return_hint, resource.Resource
    ):
        return (return_hint, False)

    origin = typing.get_origin(return_hint)
    args = typing.get_args(return_hint)

    # Generator[ResourceType, None, None]
    if origin is Generator and args:
        first = args[0]
        if isinstance(first, type) and issubclass(first, resource.Resource):
            return (first, True)

    # ResourceType | None  (types.UnionType from PEP 604)
    # typing.Union[ResourceType, None]  (legacy Optional)
    if isinstance(return_hint, types.UnionType) or origin is typing.Union:
        resource_types = [
            a
            for a in args
            if isinstance(a, type) and issubclass(a, resource.Resource)
        ]
        if len(resource_types) == 1:
            return (resource_types[0], False)

    return None


def _make_resource_side_effect(
    resource_type: type[resource.Resource],
    is_generator: bool,
) -> Any:
    """Return a callable suitable for use as a mock side_effect."""
    if is_generator:

        def _side_effect(*args: Any, **kwargs: Any) -> Any:
            return generate_fake_resources(resource_type)

    else:

        def _side_effect(*args: Any, **kwargs: Any) -> Any:
            return generate_fake_resource(resource_type)

    return _side_effect


def generate_fake_proxy(
    service: type[service_description.ServiceDescription[proxy.ProxyT]],
    api_version: str | None = None,
) -> proxy.ProxyT:
    """Generate a fake proxy for the given service type

    Proxy methods whose return type annotation is a
    :class:`~openstack.resource.Resource` subclass (or a
    ``Generator`` / optional thereof) are automatically configured to return
    fake resource instances via :func:`generate_fake_resource` or
    :func:`generate_fake_resources`.  Return values can still be overridden
    on the returned mock after the fact.

    Example usage:

    .. code-block:: python

        >>> from openstack.compute import compute_service
        >>> from openstack.compute.v2 import server
        >>> from openstack.test import fakes
        >>> fake_compute_proxy = fakes.generate_fake_proxy(
        ...    compute_service.ComputeService,
        ... )
        >>> # get_server is pre-configured; returns a fake Server
        >>> fake_server = fake_compute_proxy.get_server('some-uuid')
        >>> isinstance(fake_server, server.Server)
        True
        >>> # servers() is pre-configured; yields fake Server instances
        >>> fake_servers = list(fake_compute_proxy.servers())
        >>> isinstance(fake_servers[0], server.Server)
        True
        >>> fake_compute_proxy.serverssss()
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
          File "/usr/lib64/python3.11/unittest/mock.py", line 653, in __getattr__
            raise AttributeError("Mock object has no attribute %r" % name)
        AttributeError: Mock object has no attribute 'serverssss'. Did you mean: 'server_ips'?

    :param service: The service to generate the fake proxy for.
    :type service: :class:`~openstack.service_description.ServiceDescription`
    :param api_version: The API version to generate the fake proxy for.
        This should be a major version must be supported by openstacksdk, as
        specified in the ``supported_versions`` attribute of the provided
        ``service``. This is only required if openstacksdk supports multiple
        API versions for the given service.
    :type api_version: int or None
    :raises ValueError: if the ``service`` is not a valid
        :class:`~openstack.service_description.ServiceDescription` or if
        ``api_version`` is not supported
    :returns: An autospecced mock of the :class:`~openstack.proxy.Proxy`
        implementation for the specified service type and API version
    """  # noqa: E501
    if not issubclass(service, service_description.ServiceDescription):
        raise ValueError(
            f"Service {service.__name__} is not a valid ServiceDescription"
        )

    supported_versions = service.supported_versions

    if api_version is None:
        if len(supported_versions) > 1:
            raise ValueError(
                f"api_version was not provided but service {service.__name__} "
                f"provides multiple API versions"
            )
        else:
            api_version = next(iter(supported_versions))
    elif api_version not in supported_versions:
        raise ValueError(
            f"API version {api_version} is not supported by openstacksdk. "
            f"Supported API versions are: {', '.join(supported_versions)}"
        )

    proxy_class = supported_versions[api_version]
    fake_proxy = cast(
        proxy.ProxyT, mock.create_autospec(proxy_class, spec_set=True)
    )

    for name, _ in inspect.getmembers(
        proxy_class, predicate=inspect.isfunction
    ):
        if name.startswith('_'):
            continue
        method = getattr(proxy_class, name)
        try:
            hints = typing.get_type_hints(method)
        except Exception:  # noqa: S112
            continue
        return_hint = hints.get('return')
        if return_hint is None:
            continue
        result = _resource_type_from_hint(return_hint)
        if result is None:
            continue
        resource_type, is_generator = result
        getattr(fake_proxy, name).side_effect = _make_resource_side_effect(
            resource_type, is_generator
        )

    return fake_proxy



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Self

from keystoneauth1 import adapter

from openstack import exceptions
from openstack import resource
from openstack import utils


class ShareType(resource.Resource):
    resource_key = "share_type"
    resources_key = "share_types"
    base_path = "/types"

    # capabilities
    allow_create = True
    allow_fetch = True
    allow_commit = True
    allow_delete = True
    allow_list = True
    allow_head = False

    _query_mapping = resource.QueryParameters("is_public", "extra_specs")

    #: Properties
    #: Description of the share type
    description = resource.Body("description", type=str)
    #: List of extra specs for share types
    extra_specs = resource.Body("extra_specs", type=dict, default={})
    #: Indicates if share type is default
    is_default = resource.Body("is_default", type=bool)
    #: Indicates if share type is public
    is_public = resource.Body("share_type_access:is_public", type=bool)
    #: List of required extra specs for share types
    required_extra_specs = resource.Body("required_extra_specs", type=dict)

    def set_extra_specs(self, session: adapter.Adapter, **specs: str) -> Self:
        """Update extra_specs

        This call will replace only the extra_specs with the same keys
        given here.  Other keys will not be modified.

        :param session: The session to use for this request.
        :param extra_specs: key/value extra_specs pairs to set or update
        :returns: Updated self
        """
        url = utils.urljoin(self.base_path, self.id, 'extra_specs')
        microversion = self._get_microversion(session)
        response = session.post(
            url, json={'extra_specs': specs}, microversion=microversion
        )
        exceptions.raise_from_response(response)
        specs = response.json()['extra_specs']
        self._update(extra_specs=specs)
        return self

    def delete_extra_specs_property(
        self, session: adapter.Adapter, prop: str
    ) -> None:
        """Delete extra specs

        :param session: The session to use for this request.
        :param prop: The property to delete.
        :returns: None
        """
        url = utils.urljoin(self.base_path, self.id, "extra_specs", prop)
        microversion = self._get_microversion(session)
        response = session.delete(url, microversion=microversion)
        exceptions.raise_from_response(response)
